module 'emop.parser' {
   parse(Lexems, Mode, Fid, FidPos, Ast) :-
      module Parser [Lexems, Mode, Fid, FidPos, Parser] {
         parse_left(Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            Parser:parse_left(Lexem, Delim, Ops, Values, Ast).

         parse_right(Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            Parser:parse_right(Lexem, Delim, Ops, Values, Ast).

         parse_left(
            var(Atom, PB, Pos_end),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Values0 = [var(Atom, PB, Pos_end) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left(
            comment(Kind, Comment, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Kind = doc
            ->
               Ops0 = [op(1250, xfy, left, doc(Comment, PB, PE), PB, PE) | Ops]
            ;
               Ops0 = Ops
            ),
            Parser:parse_left(Delim, Ops0, Values, Ast).

         parse_left(
            number(Number, PB, Pos_end),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Values0 = [number(Number, PB, Pos_end) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left('('(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_left(some(')', PB, PE), [], [], V),
            Values0 = ['()'(V, PB, PE) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left('['(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_left(some(']', PB, PE), [], [], Res),
            (
               Res = callable('|', unescaped, [Seq, Tail], _, _)
            ->
               true
            ;
               Seq = Res,
               Tail = callable('[]', unescaped, [], PB, PE)
            ),
            Parser:sequence_list(Seq, List),
            Parser:append(List, Tail, PB, PE, V),
            Values0 = [V | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).
         
         parse_left('{'(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_left(some('}', PB, PE), [], [], V),
            (
               V = empty(_, _)
            ->
               Args = []
            ;
               Args = [V]
            ),
            Values0 = [callable('{}', unescaped, Args, PB, PE) | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left(
            atom(Atom, Escaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Escaped = unescaped,
               Parser:(
                  op(Atom, TrueAtom, Priority, Assoc),
                  left(Assoc)
               )
            ->
               Lexems:read(Lexem),
               (
                  Atom = '-',
                  Lexem = number(Number, _, PE0)
               ->
                  Values0 = [number('data.number':(-Number), PB, PE0) | Values],
                  Parser:parse_right(Delim, Ops, Values0, Ast)
               ;  
                  Op = op(Priority, Assoc, left, TrueAtom, PB, PE),
                  Ops0 = [Op | Ops],
                  Parser:parse_left(Lexem, Delim, Ops0, Values, Ast)
               )
            ;
               Escaped = unescaped,
               Atom = 'module'
            ->
               Parser:parse_module(PB, PE, Delim, Ops, Values, Ast)
            ;
               Escaped = unescaped
            ->
               (
                  Parser:reserved(Atom, 0)
               ->
                  Value = reserved(Atom, [], PB, PE),
                  Parser:parse_right(Delim, Ops, [Value | Values], Ast)
               ;
                  Value = callable(Atom, unescaped, [], PB, PE),
                  Parser:parse_right(Delim, Ops, [Value | Values], Ast)
               )
            ;
               Parser:parse_atom(Atom, PB, PE, Delim, Ops, Values, Ast)
            ).

         parse_left(string(String, PB, PE), Delim, Ops, Values, Ast) :-
            Parser:parse_string(String, PB, PE, Delim, Ops, Values, Ast).

         parse_atom(Atom, PB, PE, Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            (
               Lexem = end_of_line(_)
            ->
               (
                  Parser:toplevel_end_of_line(Delim, Ops, [Value | Values], Ast)
               ->
                  Value = callable(Atom, escaped, [], PB, PE)
               ;
                  Parser:parse_atom(Atom, PB, PE, Delim, Ops, Values, Ast)
               )
            ;
               Lexem = atom(Atom0, escaped, _, PE0)
            ->
               Parser:parse_atom(
                  'data.atom':concat(Atom, Atom0),
                  PB,
                  PE0,
                  Delim,
                  Ops,
                  Values,
                  Ast
               )
            ;
               Value = callable(Atom, escaped, [], PB, PE),
               Parser:parse_right(Lexem, Delim, Ops, [Value | Values], Ast)
            ).

         parse_string(String, PB, PE, Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            (
               Lexem = end_of_line(_)
            ->
               (
                  Parser:toplevel_end_of_line(Delim, Ops, [Value | Values], Ast)
               ->
                  Parser:string(String, PB, PE, Value)
               ;
                  Parser:parse_string(String, PB, PE, Delim, Ops, Values, Ast)
               )
            ;
               Lexem = string(String0, _, PE0)
            ->
               Parser:parse_string(
                  'data.list':append(String, String0),
                  PB,
                  PE0,
                  Delim,
                  Ops,
                  Values,
                  Ast
               )
            ;
               Parser:string(String, PB, PE, Value),
               Parser:parse_right(Lexem, Delim, Ops, [Value | Values], Ast)
            ).

         parse_module(PB, PE, Delim, Ops, Values, Ast) :-
            Lexems:read(Lexem),
            (
               Lexem = atom(Atom, escaped, AtomPB, AtomPE)
            ->
               (
                  'data.var':check(Fid)
               ->
                  Name = atom(Atom, AtomPB, AtomPE),
                  Fid = Atom,
                  FidPos = file(AtomPB, AtomPE)
               ;
                  throw(
                     error('File module already defined', file(PB, PE))
                  )
               ),
               Lexems:read(Lexem0)
            ;
               Lexem = var(_, _, _)
            ->
               Name = Lexem,
               Lexems:read(Lexem0)
            ;
               Name = var('_', PB, PE),
               Lexem0 = Lexem
            ),
            Parser:parse_module_body(Lexem0, Name, PB, PE, V),
            Values0 = [V | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_module_body(Lexem, Name, PB, PE, V) :-
            (
               Lexem = '{'(Curly_begin, Curly_end)
            ->
               Env = []
            ;
               Lexem = '['(Env_begin, Env_end)
            ->
               Parser:parse_left(some(']', Env_begin, Env_end), [], [], Seq),
               Parser:sequence_list(Seq, Env),
               Lexems:read(Lexem0),
               (
                  Lexem0 = '{'(Curly_begin, Curly_end)
               ->
                  true
               ;
                  throw(
                     error(
                        'Module body is missing',
                        file(PB, PE)
                     )
                  )
               )
            ),
            Parser:parse_left(some('}', Curly_begin, Curly_end), [], [], Body),
            V = 'module'(Name, Env, Body, PB, Curly_end).

         parse_left(
            functor(Atom, Escaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Parser:parse_left(some(')', PB, PE), [], [], Seq),
            Parser:sequence_list(Seq, List),
            (
               Escaped = unescaped,
               'data.list':length(List, N),
               Parser:reserved(Atom, N)
            ->
               Value = reserved(Atom, List, PB, PE)
            ;
               Value = callable(Atom, Escaped, List, PB, PE)
            ),
            Values0 = [Value | Values],
            Parser:parse_right(Delim, Ops, Values0, Ast).

         parse_left(
            close(Symbol, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Values0 = [empty(PB, PE) | Values],
            Parser:check_delim(Delim, Symbol, PB, PE),
            Parser:flush(Ops, Values0, Ast).

         parse_left(end_of_file(P), Delim, Ops, Values, Ast) :-
            (
               Mode = toplevel
            ->
               Values0 = [empty(eos, eos) | Values]
            ;
               Values0 = [empty(P, P) | Values]
            ),
            Parser:check_no_delim(Delim),
            Parser:flush(Ops, Values0, Ast).

         parse_left(end_of_line(_), Delim, Ops, Values, Ast) :-
            (
               Mode = toplevel,
               catch(
                  (
                     (
                        Ops = [op(_, _, _, '.', _, _) | _]
                     ;
                        Ops = []
                     )
                  ->
                     Values0 = [empty(eol, eol) | Values],
                     Parser:check_no_delim(Delim),
                     Parser:flush(Ops, Values0, Ast)
                  ),
                  _,
                  fail
               )
            ->
               true
            ;
               Parser:parse_left(Delim, Ops, Values, Ast)
            ).

         parse_right(
            comment(Kind, _, Pb, Pe),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Kind = 'doc'
            ->
               throw(error('Unallowed documentation here', file(Pb, Pe)))
            ;
               true
            ),
            Parser:parse_right(Delim, Ops, Values, Ast).

         parse_right(
            close(Symbol, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            Parser:check_delim(Delim, Symbol, PB, PE),
            Parser:flush(Ops, Values, Ast).

         parse_right(end_of_file(_), Delim, Ops, Values, Ast) :-
            Parser:check_no_delim(Delim),
            Parser:flush(Ops, Values, Ast).

         parse_right(end_of_line(_), Delim, Ops, Values, Ast) :-
            (
               Parser:toplevel_end_of_line(Delim, Ops, Values, Ast)
            ->
               true
            ;
               Parser:parse_right(Delim, Ops, Values, Ast)
            ).

         toplevel_end_of_line(Delim, Ops, Values, Ast) :-
            Mode = toplevel,
            catch(
               (
                  Parser:check_no_delim(Delim),
                  Parser:flush(Ops, Values, Ast)
               ),
               _,
               fail
            ).

         parse_right(
            atom(Atom, Escaped, PB, PE),
            Delim,
            Ops,
            Values,
            Ast
         ) :-
            (
               Escaped = unescaped
            ->
               true
            ;
               throw(
                  error('Operators should not be quoted', file(PB, PE))
               )
            ),
            (
               Parser:op(Atom, TrueAtom, Priority, Assoc),
               (
                  Parser:binary(Assoc)
               ->
                  Kind = binary
               ;
                  Parser:right(Assoc)
               ->
                  Kind = right
               )
            ->
               true
            ;
               throw(
                  error('Not a binary or unary right operator', file(PB, PE))
               )
            ),
            Parser:flush_upto(
               Ops,
               Values,
               PE,
               Priority,
               Assoc,
               Ops0,
               Values0
            ),
            (
               Kind = binary
            ->
               Op = op(Priority, Assoc, binary, TrueAtom, PB, PE),
               Ops1 = [Op | Ops0],
               Parser:parse_left(Delim, Ops1, Values0, Ast)
            ;
               Values0 = [Arg | Tail],
               Values1 = [callable(TrueAtom, unespaced, [Arg], PB, PE) | Tail],
               Parser:parse_right(Delim, Ops0, Values1, Ast)
            ).

         parse_right('['(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:implicit_module('['(PB, PE), Delim, Ops, Values, Ast).

         parse_right('{'(PB, PE), Delim, Ops, Values, Ast) :-
            Parser:implicit_module('{'(PB, PE), Delim, Ops, Values, Ast).

         parse_right(var(_, PB, PE), _, _, _, _) :-
            throw(
               error('Variables cannot appear as operator', file(PB, PE))
            ).

         parse_right(number(_, PB, PE), _, _, _, _) :-
            throw(
               error('Numbers cannot appear as operator', file(PB, PE))
            ).

         parse_right(functor(_, _, PB, PE), _, _, _, _) :-
            throw(
               error('Functors cannot appear as operator', file(PB, PE))
            ).

         parse_right(string(_, PB, PE), _, _, _, _) :-
            throw(
               error('Strings cannot appear as operator', file(PB, PE))
            ).

         parse_right('('(PB, PE), _, _, _, _) :-
            throw(
               error('Parentheses cannot appear as operator', file(PB, PE))
            ).

         sequence_list(empty(_, _), []) :-
            !.
         sequence_list(reserved(',', [Left, Right], _, _), [Left | Tail]) :-
            !,
            Parser:sequence_list(Right, Tail).
         sequence_list(Term, [Term]).

         append([], Tail, _, _, Tail).
         append([H | T], Tail, PB, PE, L) :-
            L = callable('.', escaped, [H, L0], PB, PE),
            Parser:append(T, Tail, PB, PE, L0).

         string([], PB, PE, callable('[]', escaped, [], PB, PE)).
         string([H | T], PB, PE, Term) :-
            Term = callable(
               '.',
               escaped,
               [callable(H, escaped, [], PB, PE), Tail],
               PB,
               PE
            ),
            Parser:string(T, PB, PE, Tail).

         eliminate_parenthesis('()'(V0, _, _), V1) :-
            !,
            Parser:eliminate_parenthesis(V0, V1).
         eliminate_parenthesis(V, V).

         implicit_module(Lexem, Delim, Ops, Values, Ast) :-
            (
               Values = [Value | Values0],
               Parser:eliminate_parenthesis(Value, Value0),
               Value0 = callable(F, Escaped, Args, PB, PE)
            ->
               Name = var('_', PB, PE),
               Parser:parse_module_body(Lexem, Name, PB, PE, V),
               'data.list':append(Args, [V], Args0),
               Values1 = [callable(F, Escaped, Args0, PB, PE) | Values0],
               Parser:parse_right(Delim, Ops, Values1, Ast)
            ;
               Lexem = '['(PB, PE)
            ->
               throw(
                  error('Brackets cannot appear as operator', file(PB, PE))
               )
            ;
               Lexem = '{'(PB, PE)
            ->
               throw(
                  error('Curly brackets cannot appear as operator', file(PB, PE))
               )
            ).

         reserved(':-', 2).
         reserved(',', 2).
         reserved('->', 2).
         reserved(';', 2).
         reserved(':', 2).
         reserved('=', 2).
         reserved('\\+', 1).
         reserved('\\=', 2).
         reserved(catch, 3).
         reserved(for, 3).
         reserved(findall, 3).
         reserved(bagof, 3).
         reserved(setof, 3).
         reserved(throw, 1).
         reserved(true, 0).
         reserved(fail, 0).
         reserved(!, 0).
         reserved(repeat, 0).
         reserved(this, 0).
         reserved(scope, 0).

         op('.', '.', 1300, xfy).
         op(':-', ':-', 1200, xfx).
         op(';', ';', 1100, xfy).
         op('->', '->', 1050, xfy).
         op('|', '|', 1025, xfy).
         op(',', ',', 1000, xfy).
         op('\\+', '\\+', 900, fy).
         op('<=>', '<=>', 750, xfy).
         op('=>', '=>', 750, xfy).
         op('\\/', '\\/', 730, yfx).
         op('/\\', '/\\', 720, yfx).
         op('\\=', '\\=', 700, xfy).
         op('=', '=', 700, xfy).
         op('<', '<', 700, xfy).
         op('>', '>', 700, xfy).
         op('=<', '=<', 700, xfy).
         op('>=', '>=', 700, xfy).
         op('..', '..', 600, xfx).
         op('+', '+', 500, yfx).
         op('-', '-', 500, yfx).
         op('*', '*', 400, yfx).
         op('/', '/', 400, yfx).
         op('mod', 'mod', 400, yfx).
         op('quo', 'quo', 400, yfx).
         op('**', '**', 200, xfy).
         op('+', '+', 200, fy).
         op('-', '-', 200, fy).
         op('++', '|++', 200, yf).
         op('--', '|--', 200, yf).
         op('++', '++|', 200, fy).
         op('--', '--|', 200, fy).
         op(':', ':', 100, yfx).

         left(fx).
         left(fy).

         binary(xfx).
         binary(xfy).
         binary(yfx).

         right(xf).
         right(yf).

         relative(xfy, xfy, succede).
         relative(yfx, yfx, precede).

         check_no_delim(none).
         check_no_delim(some(_, PB, PE)) :-
            throw(error('Unmatched delimiter', file(PB, PE))).

         check_delim(none, _, PB, PE) :-
            throw(error('Unmatched delimiter', file(PB, PE))).
         check_delim(some(Delim, _, _), Delim, _, _) :-
            !.
         check_delim(some(_, PB, _), _, _, PE) :-
            throw(
               error('Mismatched delimiters', file(PB, PE))
            ).

         apply_op(left, Atom, PB, PE, Values0, [V | Values1]) :-
            (
               Values0 = [Arg0 | Values1]
            ->
               (
                  Atom = doc(Comment, PB0, PE0)
               ->
                  V = doc(Comment, PB0, PE0, Arg0)
               ;
                  Atom = '\\+'
               ->
                  V = reserved(Atom, [Arg0], PB, PE)
               ;
                  V = callable(Atom, unescaped, [Arg0], PB, PE)
               )
            ;
               throw(
                  error('Missing operand', file(PB, PE))
               )
            ).
         apply_op(binary, Atom, PB, PE, Values0, [V | Values1]) :-
            (
               Values0 = [Arg1, Arg0 | Values1]
            ->
               (
                  Parser:reserved(Atom, 2)
               ->
                  V = reserved(Atom, [Arg0, Arg1], PB, PE)
               ;
                  V = callable(Atom, unescaped, [Arg0, Arg1], PB, PE)
               )
            ;
               throw(
                  error('Missing right operand', file(PB, PE))
               )
            ).

         flush([], Values, Ast) :-
            (
               [Ast] = Values
            ->
               true
            ;
               throw(error('flush'))
            ).
         flush(
            [doc(Comment, PB, PE) | Tail],
            Values0,
            Ast
         ) :-
            Values0 = [Value | Values1],
            Parser:flush(Tail, [doc(Comment, PB, PE, Value) | Values1], Ast).
         flush(
            [op(_, _, Kind, Op, PB, PE) | Tail],
            Values0,
            Ast
         ) :-
            Parser:(
               apply_op(Kind, Op, PB, PE, Values0, Values1),
               flush(Tail, Values1, Ast)
            ).

         flush_upto([], Values, _, _, _, [], Values).
         flush_upto(
            [Op | Tail],
            Values0,
            PE,
            Priority1,
            Assoc1,
            Ops1,
            Values1
         ) :-
            (
               Op = op(Priority0, Assoc0, Kind, Atom, PB, _),
               (
                  'data.number':(Priority0 < Priority1)
               ;
                  Priority0 = Priority1,
                  (
                     Parser:relative(Assoc0, Assoc1, Rel)
                  ->
                     (
                        Rel = precede
                     ->
                        true
                     ;
                        fail
                     )
                  ;
                     throw(error('Operator conflict', file(PB, PE)))
                  )
               )
            ->
               Parser:(
                  apply_op(Kind, Atom, PB, PE, Values0, Values),
                  flush_upto(
                     Tail,
                     Values,
                     PE,
                     Priority1,
                     Assoc1,
                     Ops1,
                     Values1
                  )
               )
            ;
               Ops1 = [Op | Tail],
               Values1 = Values0
            ).
      },
      Parser:parse_left(none, [], [], Ast).
}
